#!/usr/bin/env python3
import json
import yaml
import copy
import requests
import pathlib
import zipfile
import subprocess
import os
from mako.template import Template
from mako.lookup import TemplateLookup
from sys import exit
from shutil import copyfile, move, copytree
from mako import exceptions

class CustomEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, pathlib.PosixPath):
            return str(obj)
        return json.JSONEncoder.default(self, obj)

def custom_json_dumps(s, **kwargs):
    return json.dumps(s,cls=CustomEncoder, **kwargs)

#reference: https://github.com/samuelcolvin/pydantic/issues/473

ASCIIDOCTOR_REVEALJS_URL="https://cdnjs.cloudflare.com/ajax/libs/reveal.js/3.8.0"

OUTPUT_DIR = "{}_output"
LESSON_PLAN_TEMPLATE = "lp.template"
LESSON_PLAN_FILES_DIR = "lp_files"
LESSON_PLAN_JSON_NAME = "lp.json"
LESSON_PLAN_ADOC_NAME = "lp.adoc"
LESSON_PLAN_OUT_NAME = "index.html"
LESSON_PLAN_INSTRUCTOR_BOOK = "instructor_book.adoc"
LESSON_PLAN_STUDENT_BOOK = "student_book.adoc"
STUDENT_FILES_FOLDER_NAME = "student_files"
INSTRUCTOR_FILES_FOLDER_NAME = "instructor_files"
LESSON_SLIDES_FOLDER_NAME = "slides"
DEFAULT_SLIDE_DECK_NAME = "default_slides.adoc"
DEFAULT_SLIDE_DECK_TEMPLATE="slides.template"
DEFAULT_MTS_TEMPLATE="mts.template"
DEFAULT_MODULE_SUMMARY_TEMPLATE="module_summary.template"
MODULE_SUMMARY_MD_NAME="_index.md"
MODULE_SUMMARY_CSV_NAME="summary.csv"
MTS_MD_NAME="_index.md"
MTS_CSV_NAME="mts.csv"
TEST_FOLDER = "_test"
PE_FOLDER = "_pe"
AUTO_LESSON_SOURCE_ZIP_FN = "auto_gen_lesson_source.zip"
TRADOC_MIN_PER_HOURS = 50.0
TRADOC_HOURS_PER_WEEK = 38
TRADOC_HOURS_PER_DAY = 7.6
MTS_CSV_LABELS = [ "module_name", "start_hour", "end_hour", "title", "lesson_number", "lesson_hour","MOI_hour", "MOI", "lesson_weeks", "module_total_time"]
ANTORA_ROOT = "ROOT"
ANTORA_MOD_DIR = "ANTORA"
ANTORA_ATTACHMENTS_DIR = "attachments"
ANTORA_PAGES_DIR = "pages"
ANTORA_NAV_FILE = "nav.adoc"
HUGO_LESSON_MD_INDEX = "_index.md"
HUGO_LESSON_MD_INDEX_FN = "lesson_index.md.template"
HUGO_MENU = "index.md"
HUGO_MENU_FN = "menu.md.template"
HUGO_DIR = "HUGO"
HUGO_STATIC_DIR = "static"
HUGO_CONTENT_DIR = "content"
HUGO_STANDALONE_DIR = "static"
HUGO_GENERATED_DIR = "course"
SLIDEDECK_FN = "slidedeck_{}.pdf"
WEEKS_SIG_DIGITS = 2

MODULES_DIR = "modules"
ADMIN_FOLDER_NAME = "_info"
IGNORE_FOLDER_KEY = "_ignore"
LESSON_CONTENT_FOLDER = "_content"
MEDIA_FOLDER = "_media"
ELO_PREFIX = "_elo_"
LSA_PREFIX = "_lsa_"
ELO_KEY = "elo_list"
LSA_KEY = "lsa_list"
SUPPORTED_HUGO_TEMPLATES = [ ".hugotemplate"]
GENERIC_TEMPLATE_EXT = ".generic.template"



DEBUG_VERBOSE = 2
DEBUG_NORMAL = 1
DEBUG_LOW = 0
DEBUG_NONE = -1

MAKO_TEMPLATE_DIRECTORIES = []

if "VERBOSITY" in os.environ:
    VERBOSITY = int(os.environ["VERBOSITY"])
else:
    VERBOSITY = 2


def dbg(*msg, verbosity=DEBUG_VERBOSE):
    if verbosity <= VERBOSITY:
        print(*msg)

def err(*msg):
    dbg(*msg, verbosity=DEBUG_NONE)
    exit(1)

def add_hugo_header_to_html(path,title):
    header = """---\ntitle: {}\ntype: docs\nBookToC: false\n---\n""".format(title)
    f = pathlib.Path(path)
    text = f.read_text()
    text = header + text
    f.write_text(text)

def rewrite_file(path,old,new):
    f = pathlib.Path(path)
    text = f.read_text()
    text = text.replace(old,new)
    f.write_text(text)

def hours_mins_to_float(hours,mins,min_per_hours=TRADOC_MIN_PER_HOURS):
    return ((int(hours)*min_per_hours) + int(mins)) / min_per_hours

def escape_csv(s):
    o = str(s).replace("\n"," ").replace("\r"," ").replace(",",";")
    return o
def csv_list_to_str(l):
    output = ""
    for row in l:
        output += ",".join([escape_csv(i) for i in row]) + "\n"
    return output


unlink_queue = []

def unlink_file(f):
    if DEBUG_ONLY == False:
        unlink_queue.append(f)

def cleanup_unlink_queue():
    for f in unlink_queue:
        if f.exists():
            f.unlink()

def error_to_template():
    error = "_error.html"
    f = open(error,"wb")
    f.write(exceptions.html_error_template().render())
    f.close()
    dbg(exceptions.text_error_template().render(), verbosity=DEBUG_LOW)
    err("Failed to render. Error also saved to: {}".format(error))

def write_template(fn,d,out):
    try:
        dirs = [str(pathlib.Path(fn).parents[0])] + MAKO_TEMPLATE_DIRECTORIES
        #print("dirs is",dirs)
        thislookup = TemplateLookup(directories=dirs)
        template_fn = str(fn)
        template = Template(filename=template_fn, input_encoding='utf-8', output_encoding='utf-8', encoding_errors='replace', lookup=thislookup)
        rendered = template.render(**d)
        adoc_outfile = out
        adoc_outfile.write_bytes(rendered)
    except:
        error_to_template()


#https://gist.github.com/felixSchl/d38b455df8bf83a78d3d
def zipit(path, archname):
    dbg("zipping {} to {}".format(path,archname),verbosity=DEBUG_NORMAL)
    archive = zipfile.ZipFile(archname, "w", zipfile.ZIP_DEFLATED)
    if os.path.isdir(path):
        _zippy(path, path, archive)
    else:
        _, name = os.path.split(path)
        archive.write(path, name)
    archive.close()

def _zippy(base_path, path, archive):
    paths = os.listdir(path)
    for p in paths:
        p = os.path.join(path, p)
        if os.path.isdir(p):
            _zippy(base_path, p, archive)
        else:
            archive.write(p, os.path.relpath(p, base_path))

def find_templates(cur_file,exts=[GENERIC_TEMPLATE_EXT],recurse=True, exclude=[]):
    items = []
    if cur_file.is_dir():
        for f in cur_file.iterdir():
            if f.is_dir() and recurse:
                if f.name not in exclude:
                    items.extend(find_templates(f,exts=exts))
            elif f.is_file():
                for template in exts:
                    template_len = len(template)
                    if f.name[-template_len:] == template:
                        items.append((f,f.name[:-template_len]))
    return items

if "DEBUG_ONLY" in os.environ:
    DEBUG_ONLY = str(os.environ["DEBUG_ONLY"]) == "1" # skip decktape pdf renders
else:
    DEBUG_ONLY = False
if DEBUG_ONLY:
    dbg("disabling decktape renders", verbosity=DEBUG_LOW)

class adoc_builder:

    def run(self,cmd,input="",cwd=None):
        if isinstance(input,str):
            input = input.encode("utf-8")
        response = subprocess.run(cmd, capture_output=True,input=input,cwd=cwd)
        dbg("Running Command: {}".format(str(cmd)), verbosity=DEBUG_NORMAL)
        dbg("Output: {}".format(str(response)), verbosity=DEBUG_VERBOSE)
        return response

    def __init__(self):
        self.pdf_path = "asciidoctor-pdf"
        self.adoc_revealjs_path = "asciidoctor-revealjs"
        self.decktape_path = "decktape"
        self.adoc = "asciidoctor"
        self.antora_path = "antora"
        self.hugo_path = "/opt/appbin/hugo"


    def pdf(self,source,out_dir,stdin="", outfile="",cwd=None):
        cmd = [self.pdf_path, "-r", "asciidoctor-diagram", "-r", "asciidoctor-mathematical", "-a", "data-uri", "--safe-mode=unsafe","-D", out_dir, "--trace"]
        if outfile != "":
            cmd.append("-o")
            cmd.append(outfile)
        cmd.append(source)
        return self.run(cmd,input=stdin,cwd=cwd)

    def slides(self,source,out_dir,stdin="", outfile="index.html",cwd=None):
        cmd = [self.adoc_revealjs_path, "-r", "asciidoctor-diagram", "-r", "asciidoctor-mathematical", "-a", "data-uri", "-a", "standalone-html-mode", "--safe-mode=unsafe", "-D", out_dir, "--trace", "--backend" , "revealjs", "-a", "revealjsdir={}".format(ASCIIDOCTOR_REVEALJS_URL), "-o", outfile, source]
        return self.run(cmd,input=stdin,cwd=cwd)

    def decktape(self,source,output,stdin="",cwd=None):
        if DEBUG_ONLY:
            return
        cmd = [self.decktape_path, "--chrome-path", "chromium-browser", "--chrome-arg=--no-sandbox","--load-pause", "100", "reveal",source,output]
        return self.run(cmd,input=stdin,cwd=cwd)

    def html(self,source,out_dir,stdin="", outfile="",cwd=None,abs=True,hugo_md=False,rewrite=None):
        if abs:
            out_dir = str( pathlib.Path.cwd() / out_dir )
        if outfile == "":
            outfile = pathlib.Path(source).with_suffix(".html").name

        if hugo_md:
            outfile = pathlib.Path(outfile).with_suffix(".md")

        if cwd != None:
            outfile_full = pathlib.Path(cwd) /  pathlib.Path(out_dir) / outfile
        else:
            outfile_full = out_dir / pathlib.Path(outfile)

        outfile_stem = pathlib.Path(outfile).stem

        cmd = [self.adoc, "--safe-mode=unsafe", "-D", out_dir, "--trace"]
        if hugo_md:
            cmd.append("--no-header-footer")
            cmd.append("-a")
            cmd.append("standalone-html-mode")
        cmd += [ "--backend" , "html5", "-a", "data-uri",  "-r", "asciidoctor-diagram", "-r", "asciidoctor-mathematical", "-o", str(outfile), source]
        response = self.run(cmd,input=stdin,cwd=cwd)
        if hugo_md:
            add_hugo_header_to_html(str(outfile_full),title=outfile_stem)
            if rewrite != None:
                old = 'href="lp_files/'
                new = 'href="' + str(rewrite) + "lp_files/"
                rewrite_file(str(outfile_full),old,new)
            #rewrite URL here
        return response

    def hugo(self,cwd,stdin="",base_url=False):
        cmd = [self.hugo_path]
        if base_url:
            cmd.append("--baseURL={}".format(base_url))
        dbg("cwd: {}".format(cwd))
        response = self.run(cmd,input=stdin,cwd=cwd)
        dbg(response.stdout.decode("utf-8"),verbosity=DEBUG_LOW)
        dbg(response.stderr.decode("utf-8"),verbosity=DEBUG_LOW)
        return response



def deep_update(original,update):
    result = copy.deepcopy(original)
    for key in update.keys():
        if key in original.keys():
            if isinstance(original[key], dict) and isinstance(update[key],dict):
                result[key] = deep_update(original[key],update[key])
            elif (isinstance(original[key],list) and isinstance(update[key],list)) or (isinstance(original[key],tuple) and isinstance(update[key],tuple)):
                result[key] = original[key] + update[key]
            elif (isinstance(original[key],set) and isinstance(update[key],set)) or (isinstance(original[key],frozenset) and isinstance(update[key],frozenset)):
                result[key] = original[key] | update[key]
            else:
                result[key] = update[key]
        else:
            result[key] = update[key]
    return result



class dict_helper:

    excluded_dirs = [MEDIA_FOLDER, TEST_FOLDER, PE_FOLDER]

    def load_local_file(self,path,fn):
        dbg("load local file:{},{}".format(path,fn))
        try:
            f = pathlib.Path(pathlib.PurePath(path,fn))
            f_parents = f.parents[0]
            if f.is_file():
                return f.read_text(),f_parents
            else:
                dbg("Failed to read file: {}".format(str(f)),verbosity=DEBUG_LOW)
                return "",f_parents
        except:
            return dir,path

    def load_web_file(self,url):
        try:
            r = requests.get(url, allow_redirects=True, timeout=1)
        except:
            return url
        if r.status_code == 200:
            return r.text
        return url

    def load_local_json(self,path,d):
        try:
            text,new_dir = self.load_local_file(path,d)
            o = json.loads(text)
            return o,new_dir
        except Exception as e:
            dbg("failed to load local json",path,d,e,verbosity=DEBUG_LOW)
            return d,path

    def load_local_yaml(self,path,d):
        try:
            text,new_dir = self.load_local_file(path,d)
            o = pyyaml.safe_load(text)
            return o,new_dir
        except:
            return d,path

    def check_obj(self,o,dir):
        if isinstance(o, str):
            if o[:1] == "<":
                merger = True
            elif o[:1] == "-":
                merger = False
            else:
                return o, False

            if o[:9] == "-https://" or o[:8] == "-http://":
                return self.load_web_file(o[1:],dir),merger
            elif o[:8] == "-file://":
                return self.load_local_file(o[8:],dir)[0],merger
            elif o[1:8] == "json://":
                new_o,new_dir = self.load_local_json(o[8:],dir)
                new_o = self.walk_obj(new_o,new_dir)
                return new_o,merger
            elif o[1:8] == "yaml://":
                new_o,new_dir = self.load_local_yaml(o[8:],dir)
                new_o = self.walk_obj(new_o,new_dir)
                return new_o,merger
            else:
                return o,merger
        else:
            return self.walk_obj(o,dir),False

    def walk_obj(self,o,dir):
        if isinstance(o, list):
            new_obj = []
            for i in o:
                new_item,merger = self.check_obj(i,dir)
                if merger and isinstance(new_item,list):
                    new_obj += new_item
                else:
                    new_obj.append(new_item)
        elif isinstance(o, dict):
            new_obj = {}
            for k, v in o.items():
                new_item,merger = self.check_obj(v,dir)
                if merger and isinstance(new_item,dict):
                    new_obj = deep_update(new_obj,new_item)
                else:
                    new_obj[k] = new_item
        elif not isinstance(o,str):
            new_obj = o
        else:
            new_obj = self.check_obj(o,dir)[0]
        return new_obj

    def load_file(self,fn,fparent):

        if fn[:1] != "_":
            return {}

        if fn[-5:] == ".json":
            cur_obj,path = self.load_local_json(fparent,fn)
        elif fn[-5:] == ".yaml":
            cur_obj,path = self.load_local_yaml(fparent,fn)
        else:
                        
            cur_obj,path = self.load_local_file(fparent,fn)
            if isinstance(cur_obj,str):
                if cur_obj[:9] == "-https://" or cur_obj[:8] == "-http://":
                    cur_obj = self.load_web_file(cur_obj[1:],fparent)
                elif cur_obj[:8] == "-file://":
                    cur_obj = self.load_local_file(fparent,cur_obj[8:])[0]

        if fn[:1] == "_" and fn[:6] != "_info.":
            if fn.rfind('.') != -1: 
                fn = fn[:fn.rfind('.')] # remove extension
            cur_obj = { fn[1:] : cur_obj }
        return cur_obj

    def update_path(self,path):
        self.path = path

    def add_obj(self,obj):
        self.obj = deep_update(self.obj, obj)

    def get_obj(self):
        return copy.deepcopy(self.obj)

    def add_file(self, fn,fpath):
        new_obj = self.load_file(fn,fpath)
        new_obj = self.walk_obj(new_obj,fpath)
        if isinstance(new_obj,dict):
            self.add_obj(new_obj)

    def path_to_dict(self, dirname):
        cur_dict = {}
        p = pathlib.Path(self.path, dirname)
        if not p.is_dir():
            self.add_file(p.name,p.parent)
            return
        for folder in self.excluded_dirs:
            if dirname[0:len(folder)] == folder:
                return
        for child in p.iterdir():
            if child.is_file():
                self.add_file(child.name,child.parent)

    def __init__(self, fn="",path="",dn="",obj=False,excluded_dirs=False):
        self.path = path
        dbg(self.path)
        if excluded_dirs != False:
            self.excluded_dirs = excluded_dirs
        if obj:
            self.obj = obj
        else:
            self.obj = {}

        if fn != "":
            self.add_file(fn,path)
        elif dn != "":
            self.path_to_dict(dn)



class lsa:

    def build_lsa(self):
        self.cur_dict = dict_helper(path=self.path.parent, dn=self.path.name, obj=self.lsa_obj)

    def build_lsa_media(self):

        media_dir = self.path / MEDIA_FOLDER


        if not media_dir.is_dir():
            try:
                media_dir.mkdir()
            except:
                dbg("Error, cannot create media dir in lesson lsa path: {}\n Unable to generate books".format(str(self.path)), VERBOSITY=DEBUG_LOW)
                return
        outfile = "LSA_{}.pdf".format(self.name)
        path_outfile = self.path / MEDIA_FOLDER / outfile
        if path_outfile.is_file():
            path_outfile.unlink()
        try:
            self.lesson.module.course.adoc.pdf("-",out_dir=MEDIA_FOLDER,stdin=self.get_dict()["content"], outfile=outfile,cwd=self.path)
        except:
            dbg("error buiding media pdf for lsa {}".format(self.name), verbosity=DEBUG_LOW)
    def get_extras(self):
        pes = []
        tests = []
        for subdir in sorted(self.path.iterdir()):
            if subdir.name[0:len(TEST_FOLDER)] == TEST_FOLDER:
                test = dict_helper(path=subdir.parent, dn=subdir.name,obj={ "test_folder_name" : subdir.name }, excluded_dirs=[]).get_obj()
                tests.append(test)
            elif subdir.name[0:len(PE_FOLDER)] == PE_FOLDER:
                pe = dict_helper(path=subdir.parent, dn=subdir.name,obj={ "pe_folder_name" : subdir.name },excluded_dirs=[]).get_obj()
                pes.append(pe)
            elif subdir.name == MEDIA_FOLDER:
                self.lsa_obj["media"] = self.lesson.get_media(self.path)
                self.lesson.lesson_dict["media"].extend(self.lsa_obj["media"])
        if len(pes) > 0:
            self.lsa_obj["pes"] = pes
            self.lesson.lesson_dict["practical_exercises"].extend(pes)

        if len(tests) > 0:
            self.lsa_obj["test"] = tests
            self.lesson.lesson_dict["tests"].extend(tests)


    def get_dict(self):
        return self.cur_dict.get_obj()

    def __init__(self, lesson, path):
        self.lesson = lesson
        self.path = path
        self.name = path.name[len(LSA_PREFIX):]
        self.lsa_obj = {'lsa_folder_name' : self.name}
        self.build_lsa()
        self.build_lsa_media()
        self.get_extras()

class elo:

    def build_elo(self):

        for child in sorted(self.path.iterdir()):

            if child.name[:len(LSA_PREFIX)] == LSA_PREFIX:
                cur_lsa = lsa(self.lesson,child)
                self.cur_dict[LSA_KEY].append(cur_lsa.get_dict())

            elif child.is_file():
                elo_obj = {"elo_folder_name" : self.name}
                cur_dict = dict_helper(path=child.parent, fn=child.name, obj=elo_obj)
                self.cur_dict = deep_update(self.cur_dict, cur_dict.get_obj())

    def get_dict(self):
        return self.cur_dict

    def __init__(self, lesson, path):
        self.lesson = lesson
        self.path = path
        self.name = path.name[len(ELO_PREFIX):]
        self.cur_dict = { LSA_KEY : [] }
        self.cur_dict["media"] = self.lesson.get_media(self.path)
        self.lesson.lesson_dict["media"].extend(self.cur_dict["media"])
        self.build_elo()

class lesson:

    def calculate_academic_hours(self):
        academic_hours = {}
        LSA_HRS_LABEL = "time_of_instruction_hrs"
        LSA_MIN_LABEL = "time_of_instruction_mins"
        SUMMARY_HRS = "resident_hours"
        SUMMARY_MINS = "resident_mins"
        MOI_KEY = "method_of_instruction"
        LPI_KEY = "lesson_plan_introduction"
        SUM_KEY = "summary"
        ACADEMIC_MIN_HOURS= 50




        if LPI_KEY in self.lesson_dict.keys():
            intro = self.lesson_dict[LPI_KEY]
            if isinstance(intro,dict) and MOI_KEY in intro.keys():
                cur_moi = intro[MOI_KEY]
                if cur_moi not in academic_hours.keys():
                    academic_hours[cur_moi] = { SUMMARY_HRS: 0, SUMMARY_MINS: 0 }
                if LSA_HRS_LABEL in intro.keys():
                    academic_hours[cur_moi][SUMMARY_HRS] += int( intro[LSA_HRS_LABEL])
                if LSA_MIN_LABEL in intro.keys():
                    academic_hours[cur_moi][SUMMARY_MINS] += int( intro[LSA_MIN_LABEL])
        if SUM_KEY in self.lesson_dict.keys():
            summary = self.lesson_dict[SUM_KEY]
            if isinstance(summary,dict) and MOI_KEY in summary.keys():
                cur_moi = summary[MOI_KEY]
                if cur_moi not in academic_hours.keys():
                    academic_hours[cur_moi] = { SUMMARY_HRS: 0, SUMMARY_MINS: 0 }
                if LSA_HRS_LABEL in summary.keys():
                    academic_hours[cur_moi][SUMMARY_HRS] += int( summary[LSA_HRS_LABEL])
                if LSA_MIN_LABEL in summary.keys():
                    academic_hours[cur_moi][SUMMARY_MINS] += int( summary[LSA_MIN_LABEL])
        lsas = []
        if LSA_KEY in self.lesson_dict.keys():
            for lsa in self.lesson_dict[LSA_KEY]:
                lsas.append(lsa)
        if ELO_KEY in self.lesson_dict.keys():
            for elo in self.lesson_dict[ELO_KEY]:
                if LSA_KEY in elo.keys():
                    for lsa in elo[LSA_KEY]:
                        lsas.append(lsa)
        for lsa in lsas:
            if isinstance(lsa, dict) and MOI_KEY in lsa.keys():
                if lsa[MOI_KEY] not in academic_hours.keys():
                    academic_hours[lsa[MOI_KEY]] = { SUMMARY_HRS: 0, SUMMARY_MINS: 0 }
                if LSA_HRS_LABEL in lsa.keys():
                    academic_hours[lsa[MOI_KEY]][SUMMARY_HRS] += int(lsa[LSA_HRS_LABEL])
                if LSA_MIN_LABEL in lsa.keys():
                    academic_hours[lsa[MOI_KEY]][SUMMARY_MINS] += int(lsa[LSA_MIN_LABEL])

        final_list = []

        for method in academic_hours.keys():
            mins = academic_hours[method][SUMMARY_MINS]
            hours = academic_hours[method][SUMMARY_HRS]
            hours += int(mins/ACADEMIC_MIN_HOURS)
            mins = (mins % ACADEMIC_MIN_HOURS)
            final = {
                "academic" : "Yes",
                SUMMARY_MINS: mins,
                SUMMARY_HRS: hours,
                "methods": method
            }
            final_list.append(final)
        dbg(final_list)
        self.lesson_dict["academic_hours"] = final_list


    def make_output_dir(self):
        self.output_dir.mkdir(exist_ok = True, parents=True)
        self.output_dir_files.mkdir(exist_ok = True)
        self.output_slides_dir.mkdir(exist_ok = True)
        self.lesson_content_dir.mkdir(exist_ok = True, parents=True)


    def build_generic_slides(self):

        slides_dir = self.path / LESSON_SLIDES_FOLDER_NAME
        if not slides_dir.exists():
            slides_dir.mkdir()

        slides_output = slides_dir / DEFAULT_SLIDE_DECK_NAME

        write_template(self.get_template(template_type="slides"),self.lesson_dict,slides_output)



    # Generates asciidoctor source for a generic slide deck for the lesson
    def build_slides(self):

        slides_dir = self.path / LESSON_SLIDES_FOLDER_NAME
        if not slides_dir.exists():
            return
        if not self.output_slides_dir.is_dir():
            dbg("slides output dir does not exist!", verbosity=DEBUG_LOW)
            return
        for slidedeck in slides_dir.iterdir():
            if slidedeck.is_file() and slidedeck.suffix == ".adoc":
                deckdir = self.output_slides_dir / str(slidedeck.stem)
                deckdir.mkdir(exist_ok = True)
                output = self.module.course.adoc.slides(str(slidedeck.name),out_dir= ".",outfile="-",cwd=slides_dir)
                outfile = deckdir / "index.html"
                outfile.write_bytes(output.stdout)

                media_dir = self.path / MEDIA_FOLDER

                if not media_dir.is_dir():
                    dbg("unable to generate pdf from slides. media dir does not exist, but it should have been created by now", verbosity=DEBUG_LOW)
                    return
                slide_parent_folder = outfile.parts[::-1][1:2][0]
                slide_relative_path = LESSON_SLIDES_FOLDER_NAME / pathlib.Path(slide_parent_folder) / outfile.name
                slide_dict_entry = { "name" : str(outfile.parts[::-1][1:2][0]), "path" : str(slide_relative_path)}
                slide_fn = SLIDEDECK_FN.format(slidedeck.stem)
                slide_pdf_out = media_dir / slide_fn
                slide_input = pathlib.Path.cwd() / outfile
                self.module.course.adoc.decktape(str(slide_input),str(slide_pdf_out))
                if slide_pdf_out.is_file():
                    self.lesson_dict["media"].extend([self.add_lesson_media(slide_pdf_out)])
                    viewgraph_master = { "media_name" : slide_fn, "media_type" : "pdf", "sequence" : ""}
                    if "viewgraph_masters" not in self.lesson_dict:
                        self.lesson_dict["viewgraph_masters"] = []
                    self.lesson_dict["viewgraph_masters"].extend([viewgraph_master])
                    slide_dict_entry["pdf"] = str(slide_fn)
                    slide_dict_entry["pdf_path"] = str(slide_pdf_out)
                self.lesson_dict["slides"].append(slide_dict_entry)




    def build_appendices(self):
        student_files_list = []
        instructor_files_list = []
        #viewgraph_masters - slides
        self.build_slides()

        #TESTS should be captured in the LSAs via the _test folder

        #PEs should be captures in the LSAs in teh _pe folder

        #student handouts include student books and anything in the STUDENT_FILES_FOLDER_NAME folder
        student_book = self.path / MEDIA_FOLDER / LESSON_PLAN_STUDENT_BOOK
        student_book = student_book.with_suffix('.pdf')

        if student_book.is_file():
            student_handout = "Student Book is located in the media list under the title: {}".format(student_book.name)
            if "student_handouts" not in self.lesson_dict.keys():
                self.lesson_dict["student_handouts"] = []
            self.lesson_dict["student_handouts"].extend([student_handout])

        student_handouts_dir = self.path / MEDIA_FOLDER / STUDENT_FILES_FOLDER_NAME

        if student_handouts_dir.is_dir():
            for handout in student_handouts_dir.iterdir():
                if handout.is_dir():
                    handout_outfile = "{}.zip".format(handout)
                    zipit(str(handout),handout_outfile)
                    handout = pathlib.Path(handout_outfile)
                    self.lesson_dict["media"].extend([self.add_lesson_media(handout)])


                if handout.is_file():
                    if handout.suffix == ".adoc":
                        pdf_source = MEDIA_FOLDER / pathlib.Path(STUDENT_FILES_FOLDER_NAME) / handout.name
                        out_file = self.path / MEDIA_FOLDER / STUDENT_FILES_FOLDER_NAME / pdf_source.with_suffix('.pdf').name
                        self.module.course.adoc.pdf(str(pdf_source),str(pathlib.Path(MEDIA_FOLDER) / STUDENT_FILES_FOLDER_NAME),cwd=self.path)
                        self.lesson_dict["media"].extend([self.add_lesson_media(out_file)])
                        if self.student_mode:
                            handout.unlink()
                    else:
                        out_file = handout
                    if out_file.is_file():
                        student_handout = "Student Handout, {}, is located in the media list under the path: {}".format(out_file.name,str(out_file))
                        self.lesson_dict["student_handouts"].extend([student_handout])
                        if MEDIA_FOLDER + "/" in str(out_file):
                            relative_path= list(out_file.parts)
                            relative_index = relative_path[::-1].index(MEDIA_FOLDER) + 3
                            relative_index *= -1
                            relative_path = relative_path[relative_index:]
                            relative_path = [LESSON_PLAN_FILES_DIR] + relative_path
                            relative_path = "/".join(relative_path)
                        else:
                            relative_path = str(out_file)
                        student_files_list.append({"name" : out_file.name, "path" : relative_path})
        self.lesson_dict["student_files_list"] = student_files_list


        #trainer's lesson outline
        #Reference instructor_books.pdf, and anything in the instructor files

        if self.student_mode:
            return

        instructor_book = self.path / MEDIA_FOLDER / LESSON_PLAN_INSTRUCTOR_BOOK
        instructor_book = instructor_book.with_suffix('.pdf')

        if instructor_book.is_file():
            instructor_handout= "Trainer Lesson Outline (Instructor Book) is located in the media list under the title: {}".format(instructor_book.name)
            if "trainer_lesson_outline" not in self.lesson_dict.keys():
                self.lesson_dict["trainer_lesson_outline"] = []
            self.lesson_dict["trainer_lesson_outline"].extend([instructor_handout])

        instructor_files_dir = self.path / MEDIA_FOLDER / INSTRUCTOR_FILES_FOLDER_NAME

        if instructor_files_dir.is_dir():
            for handout in instructor_files_dir.iterdir():

                if handout.is_dir():
                    handout_outfile = "{}.zip".format(handout)
                    zipit(str(handout),handout_outfile)
                    handout = pathlib.Path(handout_outfile)
                    self.lesson_dict["media"].extend([self.add_lesson_media(handout)])

                if handout.is_file():
                    if handout.suffix == ".adoc":
                        pdf_source = MEDIA_FOLDER / pathlib.Path(INSTRUCTOR_FILES_FOLDER_NAME) / handout.name
                        out_file = self.path / MEDIA_FOLDER / INSTRUCTOR_FILES_FOLDER_NAME / pdf_source.with_suffix('.pdf').name
                        self.module.course.adoc.pdf(str(pdf_source),str(pathlib.Path(MEDIA_FOLDER) / INSTRUCTOR_FILES_FOLDER_NAME),outfile=str(out_file.name),cwd=self.path)
                        self.lesson_dict["media"].extend([self.add_lesson_media(out_file)])
                    else:
                        out_file = handout
                    if out_file.is_file():
                        if out_file.suffix == ".pdf":
                            isntructor_handout = "Instructor Resource, {}, is located in the media list under the path: {}".format(out_file.name,str(out_file))
                            self.lesson_dict["trainer_lesson_outline"].extend([instructor_handout])

                        if MEDIA_FOLDER + "/" in str(out_file):
                            relative_path= list(out_file.parts)
                            relative_index = relative_path[::-1].index(MEDIA_FOLDER) + 3
                            relative_index *= -1
                            relative_path = relative_path[relative_index:]
                            relative_path = [LESSON_PLAN_FILES_DIR] + relative_path
                            relative_path = "/".join(relative_path)
                        else:
                            relative_path = str(out_file)
                        instructor_files_list.append({"name" : out_file.name, "path" : relative_path})
        self.lesson_dict["instructor_files_list"] = instructor_files_list

    def build_zip_of_lesson(self):
        inherited_dict = self.path / "inherited_dictionary.json"
        inherited_dict.write_text(custom_json_dumps(self.lesson_dict))
        zipit(self.path,AUTO_LESSON_SOURCE_ZIP_FN)
        self.build_media_dir() 
        output = self.media_dir / AUTO_LESSON_SOURCE_ZIP_FN
        move(AUTO_LESSON_SOURCE_ZIP_FN, output)
        unlink_file(inherited_dict)

    def build_media_dir(self):
        if not self.media_dir.is_dir():
            try:
                self.media_dir.mkdir()
            except:
                err("Error, cannot create media dir in lesson path: {}".format(str(self.path)))



    def generate_books(self):

        media_dir = self.media_dir

        if not media_dir.is_dir():
            try:
                media_dir.mkdir()
            except:
                dbg("Error, cannot create media dir in lesson path: {}\n Unable to generate books".format(str(self.path)), verbosity=DEBUG_LOW)
                return

        student_book_path = self.path / LESSON_PLAN_STUDENT_BOOK
        student_book_outpath = media_dir / LESSON_PLAN_STUDENT_BOOK
        if student_book_outpath.is_file():
            unlink_file(student_book_outpath)

        if student_book_path.is_file():
            self.module.course.adoc.pdf(LESSON_PLAN_STUDENT_BOOK,MEDIA_FOLDER,cwd=self.path)
            self.module.course.adoc.html(LESSON_PLAN_STUDENT_BOOK,str(self.output_dir), cwd=self.path)
            self.module.course.adoc.html(LESSON_PLAN_STUDENT_BOOK,str(self.lesson_content_dir), cwd=self.path,hugo_md=True,rewrite=self.hugo_static_url)

        if not self.student_mode:

            instructor_book_path = self.path / LESSON_PLAN_INSTRUCTOR_BOOK
            instructor_book_outpath = media_dir / LESSON_PLAN_INSTRUCTOR_BOOK
            if instructor_book_outpath.is_file():
                unlink_file(instructor_book_outpath)


            if instructor_book_path.is_file():
                self.module.course.adoc.pdf(LESSON_PLAN_INSTRUCTOR_BOOK, MEDIA_FOLDER,cwd=self.path)
                self.module.course.adoc.html(LESSON_PLAN_INSTRUCTOR_BOOK, str(self.output_dir),cwd=self.path)
                self.module.course.adoc.html(LESSON_PLAN_INSTRUCTOR_BOOK, str(self.lesson_content_dir),cwd=self.path,hugo_md=True,rewrite=self.hugo_static_url)



    def get_template(self,template_type="lesson"):
        if template_type=="lesson":
            if hasattr(self, 'lesson_template'):
                return self.lesson_template
            elif hasattr(self.module, 'lesson_template'):
                return self.module.lesson_template
            else:
                return self.module.course.lesson_template
        elif template_type == "slides":
            if hasattr(self, 'slide_template'):
                return self.slide_template
            elif hasattr(self.module, 'slide_template'):
                return self.module.slide_template
            else:
                return self.module.course.slide_template
        elif template_type == "hugo_lesson":
            if hasattr(self, 'hugo_lesson_template'):
                return self.hugo_lesson_template
            elif hasattr(self.module, 'hugo_lesson_template'):
                return self.module.hugo_lesson_template
            else:
                return self.module.course.hugo_lesson_template

    def build_lesson_plan(self):

        #copy media over
        for media in self.lesson_dict["media"]:
            new_path = self.output_dir_files / media["lesson_relative_path"]
            if self.student_mode:
                stu_book_name = str(pathlib.Path(LESSON_PLAN_STUDENT_BOOK).with_suffix(".pdf"))
                if STUDENT_FILES_FOLDER_NAME not in str(new_path) and media["name"] != stu_book_name and not ("slidedeck_" in media["name"] and media["name"][-4:] == ".pdf"):
                    continue
            new_path.parents[0].mkdir(parents=True,exist_ok=True)
            if pathlib.Path(media["fullpath"]).exists(): #templates won't get copied.
                copyfile(media["fullpath"],str(new_path))
                dbg("copying file {} to {}".format(media["fullpath"],new_path), verbosity=DEBUG_NORMAL)

        if not self.student_mode:


        # write json to file
            json_outfile = self.output_dir / LESSON_PLAN_JSON_NAME
            json_outfile.write_text(custom_json_dumps(self.lesson_dict, sort_keys=True, indent=2, separators=(',', ': ')))

            # convert lesson to adoc
            write_template(self.get_template(),self.lesson_dict,self.output_dir / LESSON_PLAN_ADOC_NAME)
            self.module.course.adoc.html(LESSON_PLAN_ADOC_NAME,str(self.output_dir), cwd=self.output_dir)
            self.module.course.adoc.pdf(LESSON_PLAN_ADOC_NAME,".", cwd=self.output_dir)
            self.module.course.adoc.html(LESSON_PLAN_ADOC_NAME,str(self.lesson_content_dir), cwd=self.output_dir,hugo_md=True,rewrite=self.hugo_static_url)


        write_template(self.get_template(template_type="hugo_lesson"),self.lesson_dict,self.lesson_content_dir / HUGO_LESSON_MD_INDEX)

        # build lesson summary


        # copy media to out_files
        # clean up template for media

        # convert adoc to html...
        #TODO


#if folder has file called _ignore then don't parse as lessson
    def get_lesson_admin_data(self):

        self.hugo_lesson_template = self.module.hugo_lesson_template

        admin_dir = pathlib.Path(self.path, ADMIN_FOLDER_NAME)

        if admin_dir.is_dir():
            cur_dict = dict_helper(path=admin_dir.parent, dn=admin_dir.name)
            self.lesson_dict = deep_update(self.lesson_dict,cur_dict.get_obj())

            lesson_template = admin_dir / self.module.course.lesson_template_name
            slide_template = admin_dir / self.module.course.slide_template_name
            hugo_lesson_template = admin_dir / self.module.course.hugo_lesson_template_name

            if lesson_template.exists() and lesson_template.is_file():
                self.lesson_template = lesson_template
            else:
                self.lesson_template = self.module.lesson_template

            if slide_template.exists() and slide_template.is_file():
                self.slide_template = slide_template
            else:
                self.slide_template = self.module.slide_template

            if hugo_lesson_template.exists() and hugo_lesson_template.is_file():
                self.hugo_lesson_template = hugo_lesson_template

    def get_lesson_content(self):

        content_dir = pathlib.Path(self.path, LESSON_CONTENT_FOLDER)

        if content_dir.is_dir():

            for child in sorted(content_dir.iterdir()):

                if child.name[:len(LSA_PREFIX)] == LSA_PREFIX:
                    cur = lsa(self,path=child)
                    self.lesson_dict[LSA_KEY].append(cur.get_dict())

                elif child.name[:len(ELO_PREFIX)] == ELO_PREFIX:
                    cur = elo(self,path=child)
                    self.lesson_dict[ELO_KEY].append(cur.get_dict())

    def get_media(self,cur_dir,recurse=False):
        media_list = []
        if cur_dir.name == MEDIA_FOLDER or recurse:
            for file in cur_dir.iterdir():
                if file.is_file():
                #if file.name[-len(GENERIC_TEMPLATE_EXT):] != GENERIC_TEMPLATE_EXT:
                    media_list.extend([self.add_lesson_media(file)])
                elif file.is_dir():
                    media_list.extend(self.get_media(file,recurse=True))
        else:
            for child in cur_dir.iterdir():
                if child.is_dir():
                    media_list.extend(self.get_media(child))
        return media_list

    def unlink_file(self, template):
        if False == DEBUG_ONLY:
            self.unlink_queue.append(template)

    def unlink_lesson_files(self):
        for unlink_file in self.unlink_queue:
            if unlink_file.exists():
                unlink_file.unlink()

    def convert_generic_templates(self):

        media_templates = find_templates(self.path)

        for template,new_fn in media_templates:
            out_fn = template.with_name(new_fn)
            write_template(template,self.lesson_dict,out_fn)
        for template,new_fn in media_templates:
            self.unlink_file(template)

    def process_hugo_templates(self):

        generic_templates = find_templates(self.path, exts=SUPPORTED_HUGO_TEMPLATES,recurse=False)

        for template,new_fn in generic_templates:
            out_fn = self.lesson_content_dir / new_fn
            write_template(template,self.lesson_dict,out_fn)

    def add_lesson_media(self,f):
        new_media = {}
        new_media["fullpath"] = str(f)
        new_media["name"] = f.name
        new_media["parent"] = str(f.parent)
        new_media["lesson_relative_path"] = "/".join(f.parts[3:])
        return new_media

    def set_lesson_title(self):
        if "lesson_title" not in self.lesson_dict.keys():
            self.lesson_dict["lesson_title"] = self.lesson_folder

    def __init__(self,module, lesson_folder):
        self.module = module
        self.lesson_folder = lesson_folder
        self.student_mode = self.module.course.student_mode
        self.path = pathlib.Path(self.module.path, self.lesson_folder)
        self.output_dir = self.module.output_dir / self.lesson_folder
        self.lesson_content_dir = self.module.module_content_dir / self.lesson_folder
        self.output_dir_files = self.output_dir / LESSON_PLAN_FILES_DIR
        self.output_slides_dir = self.output_dir / LESSON_SLIDES_FOLDER_NAME
        self.hugo_static_url = self.module.hugo_static_url + "/" + self.lesson_folder + "/"
        self.lesson_dict = copy.deepcopy(self.module.module_dict)
        self.lesson_dict["media"] = []
        self.lesson_dict["slides"] = []
        self.media_dir = self.path / MEDIA_FOLDER
        self.unlink_queue = []
        if LSA_KEY not in self.lesson_dict:
            self.lesson_dict[LSA_KEY] = []

        if ELO_KEY not in self.lesson_dict:
            self.lesson_dict[ELO_KEY] = []

        self.lesson_dict["practical_exercises"] = []
        self.lesson_dict["tests"] = []
        self.lesson_dict["lesson_folder"] = str(self.lesson_folder)

        dbg("building lesson {} ".format(self.lesson_folder), verbosity=DEBUG_NORMAL)

        self.build_media_dir()

        if not self.student_mode:
            self.build_zip_of_lesson()
        

        self.make_output_dir()

        self.get_lesson_admin_data()
    
        self.get_lesson_content()

        self.convert_generic_templates()

        self.generate_books()

        self.unlink_lesson_files()

        self.lesson_dict["media"].extend(self.get_media(self.path / MEDIA_FOLDER))
        self.lesson_dict["hugo_lesson_static_url"]  = self.hugo_static_url

        self.calculate_academic_hours()
        self.set_lesson_title()
        self.build_generic_slides()
        self.build_appendices()


class module:

    def write_module_summary(self):
        summary_outfile = self.module_hugo_content_dir / MODULE_SUMMARY_MD_NAME

        summary_dict = self.module_dict
        summary_dict["mts_summary"] = self.module_summary
        summary_dict["mts_csv_list"] = self.module_summary_csv_list

        write_template(self.module_summary_template,summary_dict,summary_outfile)

        summary_csv = [MTS_CSV_LABELS] + self.module_summary_csv_list
        summary_csv_outfile = self.output_dir / MODULE_SUMMARY_CSV_NAME
        summary_csv_outfile.write_text(csv_list_to_str(summary_csv))



    #REFERENCE MTS HERE
    def build_module_summary(self, module_hours_start=0):

        cur_hours_start = module_hours_start
        cur_hours_end = 0
        summary_list = []
        day_list_lessons = []
        excess = 0

        for name in sorted(self.lessons, key=lambda x: self.lessons[x].lesson_folder):
            total_hours = 0
            total_weeks = 0.0
            moi_list = []
            l = self.lessons[name].lesson_dict
            summary =  {
                "cumulative_hours_start": cur_hours_start
            }
            try:
                summary["lesson_title"] = l["lesson_title"]
            except:
                summary["lesson_title"] = ""

            try:
                summary["lesson_number"] = l["lesson_number"]
            except:
                summary["lesson_number"] = ""
            try:
                cur_hours = hours_mins_to_float(l["lesson_plan_introduction"]["time_of_instruction_hrs"], l["lesson_plan_introduction"]["time_of_instruction_mins"])
                moi_entry = {
                    "title" : "Introduction - TLO: {}".format(l["lesson_plan_introduction"]["terminal_learning_objective"]),
                    "hours" : cur_hours,
                    "moi" : l["lesson_plan_introduction"]["method_of_instruction"]
                    }
                moi_list.append(moi_entry)
                total_hours += cur_hours
            except:
                pass

            for lsa in l["lsa_list"]:
                try:
                    cur_hours = hours_mins_to_float(lsa["time_of_instruction_hrs"], lsa["time_of_instruction_mins"])
                    moi_entry = {
                        "title" : "LSA - {}".format(lsa["title"]),
                        "hours" : cur_hours,
                        "moi" : lsa["method_of_instruction"]
                    }
                    moi_list.append(moi_entry)
                    total_hours += cur_hours
                except:
                    continue
            try:
                cur_hours = hours_mins_to_float(l["summary"]["time_of_instruction_hrs"], l["summary"]["time_of_instruction_mins"])
                moi_entry = {
                    "title" : "Summary",
                    "hours" : cur_hours,
                    "moi" : l["summary"]["method_of_instruction"]
                    }
                moi_list.append(moi_entry)
                total_hours += cur_hours
            except:
                pass

            summary["moi_list"] = moi_list
            cur_hours_end = cur_hours_start + total_hours
            summary["cumulative_hours_end"] = cur_hours_end
            summary["total_hours"] = total_hours
            summary["total_weeks"] = total_hours / (TRADOC_HOURS_PER_WEEK)
            day_list_lessons_days = int( ( total_hours + excess) / (TRADOC_HOURS_PER_DAY))
            if day_list_lessons_days > 0:
                excess = (total_hours + excess) % ( day_list_lessons_days * (TRADOC_HOURS_PER_DAY) )
                day_list_lessons += [summary["lesson_title"]] * day_list_lessons_days
            else:
                excess += total_hours
            summary_list.append(summary)
            cur_hours_start =cur_hours_end

        self.day_list_lessons = day_list_lessons
        self.module_summary = summary_list
        self.module_total_hours = cur_hours_end - module_hours_start
        self.module_summary_to_csv_list()
        self.write_module_summary()
        return cur_hours_end

    def module_summary_to_csv_list(self):
        summary_list = []
        #summary_list_heading = [ "module_name", "start_hour", "end_hour", "title", "lesson_number", "lesson_hour","MOI_hour", "MOI", "lesson_weeks", "module_total_time"]
        module_row = [ str(self.module_name), "", "", "", "", "","", "", "", ""]
        summary_list.append(module_row)
        for lesson in self.module_summary:
            lesson_row = [ "", lesson["cumulative_hours_start"], lesson["cumulative_hours_end"], lesson["lesson_title"], lesson["lesson_number"], lesson["total_hours"],"", "", round(lesson["total_weeks"],WEEKS_SIG_DIGITS), ""]
            summary_list.append(lesson_row)
            for moi in lesson["moi_list"]:
                moi_row =  [ "", "", "", moi["title"], "", "",moi["hours"], moi["moi"], "", ""]
                summary_list.append(moi_row)
        module_row = [ "", "", "", "", "", "","", "", "", self.module_total_hours]
        summary_list.append(module_row)
        self.module_summary_csv_list = summary_list

    def make_output_dir(self):

        if False == True:
            if self.module_antora_dir.is_dir():
                copytree(self.module_antora_dir, self.output_dir)
            else:
                dbg("antora modules folder not found. skipping:{}".format(self.module_antora_dir))
                self.output_dir.mkdir(exist_ok = True, parents=True)
            self.module_antora_attachments.mkdir(exist_ok = True, parents = True)
            self.module_antora_pages.mkdir(exist_ok = True)
        else:

            self.module_hugo_content_dir = self.module_content_dir
            self.output_dir.mkdir(exist_ok = True, parents=True)
            self.module_content_dir.mkdir(exist_ok = True, parents=True)




    def build_lessons_plans(self):
        for name in self.lessons.keys():
            self.lessons[name].build_lesson_plan()
            self.lessons[name].process_hugo_templates()
        self.process_generic_templates()
        self.process_hugo_templates()

    def get_module_admin_data(self):

        self.lesson_template = self.course.lesson_template
        self.slide_template = self.course.slide_template
        self.module_summary_template = self.course.module_summary_template
        self.hugo_lesson_template = self.course.hugo_lesson_template


        admin_dir = pathlib.Path(self.path, ADMIN_FOLDER_NAME)

        if admin_dir.is_dir():
            cur_dict = dict_helper(path=admin_dir.parent, dn=admin_dir.name)
            self.module_dict = deep_update(self.module_dict,cur_dict.get_obj())

            lesson_template = admin_dir / self.course.lesson_template_name
            slide_template = admin_dir / self.course.slide_template_name
            module_summary_template = admin_dir / self.course.module_summary_template_name
            hugo_lesson_template = admin_dir / self.course.lesson_template_name


            if lesson_template.exists() and lesson_template.is_file():
                self.lesson_template = lesson_template


            if slide_template.exists() and slide_template.is_file():
                self.slide_template = slide_template


            if module_summary_template.exists() and module_summary_template.is_file():
                self.module_summary_template = module_summary_template

            if hugo_lesson_template.exists() and hugo_lesson_template.is_file():
                self.hugo_lesson_template = hugo_lesson_template


    def iterate_lessons(self):
        lesson_dict_list = []
        p = self.path
        if not p.is_dir():
            return

        for child in p.iterdir():
            if child.name == ANTORA_MOD_DIR:
                continue

            if child.is_dir():
                if not child.name == ADMIN_FOLDER_NAME and not pathlib.Path(child, IGNORE_FOLDER_KEY).is_file():
                    self.lessons[child.name] = lesson(self,child.name)
                    lesson_dict_list.append(self.lessons[child.name].lesson_dict)
        self.module_dict["lessons"] = sorted(lesson_dict_list, key=lambda x: x["lesson_folder"])


    def process_generic_templates(self):

        media_templates = find_templates(self.path, recurse=False)

        for template,new_fn in media_templates:
            out_fn = self.output_dir / template.with_name(new_fn).name
            write_template(template,self.module_dict,out_fn)
        for template,new_fn in media_templates:
            unlink_file(template)



    def process_hugo_templates(self):

        generic_templates = find_templates(self.path, exts=SUPPORTED_HUGO_TEMPLATES,recurse=False)

        for template,new_fn in generic_templates:
            out_fn = self.module_content_dir / new_fn
            write_template(template,self.module_dict,out_fn)


    def __init__(self,parent_course, module_folder):
        self.course = parent_course
        self.module_folder = pathlib.Path(module_folder)
        self.module_dict = copy.deepcopy(self.course.course_dict)
        self.module_dict["module_folder"] = str(module_folder)

        self.lessons = {}
        self.path = pathlib.Path(self.course.course_dir, self.course.modules_dir, self.module_folder)
        self.module_antora_dir = self.path / ANTORA_MOD_DIR
        self.module_antora_attachments = self.module_antora_dir / ANTORA_ATTACHMENTS_DIR
        self.module_antora_pages = self.module_antora_dir /ANTORA_PAGES_DIR
        self.output_dir = self.course.output_dir / HUGO_STANDALONE_DIR / self.module_folder
        self.module_nav = self.output_dir / ANTORA_NAV_FILE
        self.hugo_static_url = self.course.base_url + HUGO_STANDALONE_DIR + "/" + module_folder
        self.module_content_dir = self.course.hugo_content_output / HUGO_GENERATED_DIR / module_folder
        self.make_output_dir()
        self.get_module_admin_data()
        if "module_name" in self.module_dict.keys():
            self.module_name = self.module_dict["module_name"]
        else:
            self.module_name = self.module_folder
            self.module_dict["module_name"] = str(self.module_name)
        dbg("building module {} ".format(self.module_folder), verbosity=DEBUG_LOW)
        self.iterate_lessons()


class course:

    def make_output_dir(self):
        hugo_dir = self.output_dir / HUGO_STATIC_DIR
        output_modules_dir =  hugo_dir / MODULES_DIR
        antora_ROOT_output = output_modules_dir / ANTORA_ROOT


        if self.hugo_dir.is_dir():
            try:
                if self.output_dir.is_dir():
                    self.output_dir.rmdir()
                copytree(self.hugo_dir,self.output_dir)
            except:
                err("unable to create output_dir from hugo directory")

        try:
            output_modules_dir.mkdir(exist_ok = True, parents=True)
        except FileExistsError:
            err("Cannot output, {} is a file, not a directory. Quitting.".format(self.output_dir))

        self.top_output_dir = self.output_dir
        self.output_dir = hugo_dir
        self.hugo_content_output = self.top_output_dir / HUGO_CONTENT_DIR

        return
        self.antora_ROOT_attachments = antora_ROOT_output / ANTORA_ATTACHMENTS_DIR
        self.antora_ROOT_pages = antora_ROOT_output / ANTORA_PAGES_DIR
        if self.antora_ROOT_module.is_dir():
            try:
                copytree(self.antora_ROOT_module,antora_ROOT_output)
            except:
                err("Error copying antora ROOT module")
        else:
            dbg("no antora root module found. skipping: {}".format(self.antora_ROOT_module))

        self.antora_ROOT_attachments.mkdir(exist_ok = True, parents=True)
        self.antora_ROOT_pages.mkdir(exist_ok = True)



    def finalize_mts(self):
        empty_row = [""]*len(MTS_CSV_LABELS)
        for i in range(3):
            self.mts.append(empty_row)

        course_total_ach_hours_row = empty_row.copy()
        course_total_ach_hours_row[0] = "Course Total ACH Hours"
        course_total_ach_hours_row[1] = self.course_dict["course_total_ach"]
        self.mts.append(course_total_ach_hours_row)

        course_total_weeks_row = empty_row.copy()
        course_total_weeks_row[0] = "Course Total Weeks"
        course_total_weeks_row[1] = self.course_dict["course_total_weeks"]
        self.mts.append(course_total_weeks_row)

        course_total_admin_row = empty_row.copy()
        course_total_admin_row[0] = "Course Total Admin Hours"
        course_total_admin_row[1] = self.course_dict["course_total_admin"]
        self.mts.append(course_total_admin_row)

        course_total_length_row = empty_row.copy()
        course_total_length_row[0] = "Course Total Length (Hours)"
        course_total_length_row[1] = self.course_dict["course_total_length"]
        self.mts.append(course_total_length_row)

        self.mts_csv = csv_list_to_str(self.mts)

    def write_summary(self):
        summary_outfile = self.hugo_content_output/ MTS_MD_NAME

        summary_dict = self.course_dict
        summary_dict["mts"] = self.mts

        write_template(self.mts_template,summary_dict,summary_outfile)


        summary_csv_outfile = self.output_dir/ MTS_CSV_NAME
        summary_csv_outfile.write_text(self.mts_csv)

    def write_menu(self):
        menu_outfile = self.hugo_content_output / "menu" / HUGO_MENU
        write_template(self.hugo_menu_template, self.course_dict,menu_outfile)

    def build_hugo(self):
        base_url = self.base_url
        if self.base_host:
            base_url = self.base_host + base_url
        self.adoc.hugo(self.top_output_dir,base_url=base_url)

    def process_hugo_templates(self):

        generic_templates = find_templates(self.top_output_dir, exts=SUPPORTED_HUGO_TEMPLATES,recurse=True)

        for template,new_fn in generic_templates:
            out_fn = template.with_name(new_fn)
            write_template(template,self.course_dict,out_fn)
        for template,new_fn in generic_templates:
            unlink_file(template)

    def process_generic_templates(self):

        media_templates = find_templates(self.hugo_static_dir, recurse=False)

        for template,new_fn in media_templates:
            out_fn = template.with_name(new_fn)
            write_template(template,self.course_dict,out_fn)
        for template,new_fn in media_templates:
            unlink_file(template)


    def convert_static_adoc_to_pdf(self):
        adocs_for_conversion = find_templates(self.hugo_static_dir, exts=[".adoc"], recurse=True,exclude=["static"])

        for path,fn_root in adocs_for_conversion:
            self.adoc.pdf(str(path.name),".",cwd=path.parents[0])
        for path,fn_root in adocs_for_conversion:
            unlink_file(path)

    def save_course_dict(self):
        course_dict_outfile = self.output_dir / "course_dict.json"
        course_dict_outfile.write_text(custom_json_dumps(self.course_dict))

    def build_modules(self):
        hours = 0.0
        day_list_module = []
        day_list_lesson = []
        excess = 0

        for name in sorted(self.modules, key=lambda x: self.modules[x].module_folder):
            self.modules[name].build_lessons_plans()
            hours = self.modules[name].build_module_summary(hours)
            module_days = int( (hours + excess)/ (TRADOC_HOURS_PER_DAY))
            if module_days > 0:
                excess = (hours + excess) % ( module_days * (TRADOC_HOURS_PER_DAY) )
                day_list_module += [self.modules[name].module_name] * module_days
            else:
                excess += hours
            day_list_lesson += self.modules[name].day_list_lessons
            self.mts += self.modules[name].module_summary_csv_list

        self.course_dict["poi_module_day_list"] = day_list_module
        self.course_dict["poi_lesson_day_list"] = day_list_lesson

        self.course_dict["course_total_ach"] = hours
        self.course_dict["course_total_weeks"] = round(hours / TRADOC_HOURS_PER_WEEK, WEEKS_SIG_DIGITS)
        self.course_dict["course_total_admin"] = self.course_dict["course_total_weeks"] * 2
        self.course_dict["course_total_length"] = self.course_dict["course_total_ach"] + self.course_dict["course_total_admin"]

        self.finalize_mts()
        self.write_summary()
        self.write_menu()

        if not self.student_mode:
            self.save_course_dict()



    def get_course_admin_data(self):

        self.lesson_template = pathlib.Path(self.lesson_template_name)
        self.slide_template = pathlib.Path(self.slide_template_name)
        self.mts_template = pathlib.Path(self.mts_template_name)
        self.module_summary_template = pathlib.Path(self.module_summary_template_name)
        self.hugo_menu_template  = pathlib.Path(self.hugo_menu_template_name)
        self.hugo_lesson_template = pathlib.Path(self.hugo_lesson_template_name)

        admin_dir = pathlib.Path(self.course_dir, ADMIN_FOLDER_NAME)

        if admin_dir.is_dir():
            cur_dict = dict_helper(path=admin_dir.parent, dn=admin_dir.name)
            self.course_dict = deep_update(self.course_dict,cur_dict.get_obj())

            lesson_template = admin_dir / self.lesson_template_name
            slide_template = admin_dir / self.slide_template_name
            mts_template = admin_dir / self.mts_template_name
            module_summary_template = admin_dir / self.module_summary_template_name
            hugo_menu_template = admin_dir / self.hugo_menu_template_name
            hugo_lesson_template = admin_dir / self.hugo_lesson_template_name



            if lesson_template.exists() and lesson_template.is_file():
                self.lesson_template = lesson_template
            if slide_template.exists() and slide_template.is_file():
                self.slide_template = slide_template
            if mts_template.exists() and mts_template.is_file():
                self.mts_template = mts_template
            if module_summary_template.exists() and module_summary_template.is_file():
                self.module_summary_template = module_summary_template
            if hugo_menu_template.exists() and hugo_menu_template.is_file():
                self.hugo_menu_template = hugo_menu_template
            if hugo_lesson_template.exists() and hugo_lesson_template.is_file():
                self.hugo_lesson_template = hugo_lesson_template


        if not self.lesson_template.exists():
            err("Unable to find a lesson template: {}".format(self.lesson_template))
        if not self.slide_template.exists():
            err("Unable to find a slide template: {}".format(self.slide_template))
        if not self.mts_template.exists():
            err("Unable to find an mts template: {}".format(self.mts_template))
        if not self.module_summary_template.exists():
            err("Unable to find a module summary template: {}".format(self.module_summary_template))
        if not self.hugo_menu_template.exists():
            err("Unable to find navigation menu template: {}".format(self.hugo_menu_template))
        if not self.hugo_lesson_template.exists():
            err("Unable to find navigation menu template: {}".format(self.hugo_lesson_template))


    def iterate_modules(self):

        modules_dict_list = []
        p = pathlib.Path(self.course_dir, self.modules_dir)
        if not p.is_dir():
            return

        for child in p.iterdir():
            if child.name == ANTORA_ROOT:
                continue
            if child.is_dir():
                if not pathlib.Path(child, IGNORE_FOLDER_KEY).is_file():
                    self.modules[child.name] = module(self,child.name)
                    modules_dict_list.append(self.modules[child.name].module_dict)
        modules_dict_list = sorted(modules_dict_list, key=lambda x: x["module_folder"])
        self.course_dict["modules"] = modules_dict_list

    def __init__(self,course, modules_dir=MODULES_DIR, output_dir=OUTPUT_DIR, lesson_template = LESSON_PLAN_TEMPLATE, slide_template = DEFAULT_SLIDE_DECK_TEMPLATE, mts_template = DEFAULT_MTS_TEMPLATE, module_summary_template = DEFAULT_MODULE_SUMMARY_TEMPLATE, hugo_menu_template = HUGO_MENU_FN, hugo_lesson_template = HUGO_LESSON_MD_INDEX_FN, base_url="./", base_host=False, student_mode=False):
        self.base_url = base_url
        self.base_host = base_host
        self.student_mode = student_mode
        self.modules_dir = modules_dir
        self.course_dir = course
        self.antora_ROOT_module = pathlib.Path(self.course_dir) / pathlib.Path(self.modules_dir) / ANTORA_ROOT
        self.hugo_dir = pathlib.Path(self.course_dir) / HUGO_DIR
        self.hugo_menu_template_name = hugo_menu_template
        self.hugo_lesson_template_name = hugo_lesson_template
        self.lesson_template_name = lesson_template
        self.slide_template_name = slide_template
        self.mts_template_name = mts_template
        self.module_summary_template_name = module_summary_template
        self.course_dict = { "lesson_plan_files_dir" : LESSON_PLAN_FILES_DIR, "base_url" : self.base_url, "student_mode" : self.student_mode, "media_dir" : MEDIA_FOLDER }
        self.modules = {}
        self.mts = [MTS_CSV_LABELS]
        self.adoc = adoc_builder()
        if '{}' in output_dir:
            self.output_dir = output_dir.format(course)
        else:
            self.output_dir = output_dir
        dbg("building course {} -- student mode: {}".format(self.course_dir,self.student_mode), verbosity=DEBUG_LOW)
        self.output_dir = pathlib.Path(self.output_dir)
        MAKO_TEMPLATE_DIRECTORIES.append(str(self.output_dir))
        MAKO_TEMPLATE_DIRECTORIES.append(str(self.course_dir))
        self.hugo_static_dir = self.output_dir / "static"
        self.make_output_dir()
        self.get_course_admin_data()
        self.iterate_modules()
        self.build_modules()
        self.process_generic_templates()
        self.convert_static_adoc_to_pdf()
        self.process_hugo_templates()
        cleanup_unlink_queue()
        self.build_hugo()

"""
TODO:
- map output into lesson plan builder directly X
  -> merge process template and process lesson X
  -> make a method in course such as "build modules" X
  -> make a method in the module such as "build lessons plans" X
  -> make a method in the lesson such as "build lesson plan" (integrate process template code here) X
- make lesson plan template fault tolerant
- build example course describing how to use the tool
- lesson, module, and course should auto-build a summary
- lessons should auto-build slides
- break it down into deployment phases.
"""
